# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2020) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""Enforces minimum distances on coordinates within discrete distribtution.

Copyright 2020 IMTEK Simulation
University of Freiburg

Authors:

    Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>

Examples
-------
Benchmark different scipy optimizers for the steric correction problem:

    >>> # measures of box
    >>> xsize = ysize = 5e-9 # nm, SI units
    >>> zsize = 10e-9    # nm, SI units
    >>>
    >>> # get continuum distribution, z direction
    >>> x = np.linspace(0, zsize, 2000)
    >>> c = [0.1,0.1]
    >>> z = [1,-1]
    >>> u = 0.05
    >>>
    >>> phi = potential(x, c, z, u)
    >>> C   = concentration(x, c, z, u)
    >>> rho = charge_density(x, c, z, u)
    >>>
    >>> # create distribution functions
    >>> distributions = [interpolate.interp1d(x,c) for c in C]
    >>>
    >>> # sample discrete coordinate set
    >>> box = np.array([xsize, ysize, zsize])m
    >>> sample_size = 100
    >>>
    >>> samples = [ continuous2discrete(
    >>>     distribution=d, box=box, count=sample_size) for d in distributions ]
    >>>
    >>> # apply penalty for steric overlap
    >>> x = np.vstack(samples)
    >>>
    >>> box = np.array([[0.,0.,0],box]) # needs lower corner
    >>>
    >>> n = x.shape[0]
    >>> dim = x.shape[1]
    >>>
    >>> # benchmakr methods
    >>> mindsq, (p1,p2) = scipy_distance_based_closest_pair(x)
    >>> pmin = np.min(x,axis=0)
    >>> pmax = np.max(x,axis=0)
    >>> mind = np.sqrt(mindsq)
    >>> logger.info("Minimum pair-wise distance in sample: {}".format(mind))
    >>> logger.info("First sample point in pair:    ({:8.4e},{:8.4e},{:8.4e})".format(*p1))
    >>> logger.info("Second sample point in pair    ({:8.4e},{:8.4e},{:8.4e})".format(*p2))
    >>> logger.info("Box lower boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box[0]))
    >>> logger.info("Minimum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmin))
    >>> logger.info("Maximum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmax))
    >>> logger.info("Box upper boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box[1]))
    >>>
    >>> # stats: method, x, res, dt, mind, p1, p2 , pmin, pmax
    >>> stats = [('initial',x,None,0,mind,p1,p2,pmin,pmax)]
    >>>
    >>> r = 4e-10 # 4 Angstrom steric radius
    >>> logger.info("Steric radius: {:8.4e}".format(r))
    >>>
    >>> methods = [
    >>>     'Powell',
    >>>     'CG',
    >>>     'BFGS',
    >>>     'L-BFGS-B'
    >>> ]
    >>>
    >>> for m in methods:
    >>>     try:
    >>>         logger.info("### {} ###".format(m))
    >>>         t0 = time.perf_counter()
    >>>         x1, res = apply_steric_correction(x,box=box,r=r,method=m)
    >>>         t1 = time.perf_counter()
    >>>         dt = t1 - t0
    >>>         logger.info("{} s runtime".format(dt))
    >>>
    >>>         mindsq, (p1,p2) = scipy_distance_based_closest_pair(x1)
    >>>         mind = np.sqrt(mindsq)
    >>>         pmin = np.min(x1,axis=0)
    >>>         pmax = np.max(x1,axis=0)
    >>>
    >>>         stats.append([m,x1,res,dt,mind,p1,p2,pmin,pmax])
    >>>
    >>>         logger.info("Minimum pair-wise distance in final configuration: {:8.4e}".format(mind))
    >>>         logger.info("First sample point in pair:    ({:8.4e},{:8.4e},{:8.4e})".format(*p1))
    >>>         logger.info("Second sample point in pair    ({:8.4e},{:8.4e},{:8.4e})".format(*p2))
    >>>         logger.info("Box lower boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box[0]))
    >>>         logger.info("Minimum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmin))
    >>>         logger.info("Maximum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmax))
    >>>         logger.info("Box upper boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box[1]))
    >>>     except:
    >>>         logger.warn("{} failed.".format(m))
    >>>         continue
    >>>
    >>> stats_df = pd.DataFrame( [ {
    >>>     'method':  s[0],
    >>>     'runtime': s[3],
    >>>     'mind':    s[4],
    >>>     **{'p1{:d}'.format(i): c for i,c in enumerate(s[5]) },
    >>>     **{'p2{:d}'.format(i): c for i,c in enumerate(s[6]) },
    >>>     **{'pmin{:d}'.format(i): c for i,c in enumerate(s[7]) },
    >>>     **{'pmax{:d}'.format(i): c for i,c in enumerate(s[8]) }
    >>> } for s in stats] )
    >>>
    >>> print(stats_df.to_string(float_format='%8.6g'))
    method  runtime        mind         p10         p11         p12         p20         p21         p22       pmin0       pmin1       pmin2       pmax0       pmax1       pmax2
    0   initial        0 1.15674e-10 2.02188e-09 4.87564e-10 5.21835e-09 2.03505e-09 3.72691e-10 5.22171e-09 1.17135e-12 1.49124e-10 6.34126e-12 4.98407e-09 4.99037e-09 9.86069e-09
    1    Powell  75.2704 8.02318e-10 4.23954e-09 3.36242e-09 8.80092e-09 4.31183e-09 2.56345e-09 8.81278e-09 4.01789e-10  4.0081e-10  4.2045e-10 4.59284e-09 4.54413e-09  9.5924e-09
    2        CG  27.0756  7.9992e-10 3.39218e-09 4.00079e-09 8.27255e-09 3.86337e-09 4.27807e-09 7.68863e-09 4.00018e-10 4.00146e-10 4.00565e-10 4.59941e-09 4.59989e-09 9.59931e-09
    3      BFGS  19.0255 7.99527e-10 1.82802e-09 3.54397e-09 9.69736e-10 2.41411e-09   3.936e-09 1.34664e-09 4.00514e-10 4.01874e-10  4.0002e-10 4.59695e-09 4.59998e-09 9.58155e-09
    4  L-BFGS-B  11.7869 7.99675e-10 4.34395e-09 3.94096e-09 1.28996e-09 4.44064e-09 3.15999e-09 1.14778e-09 4.12146e-10 4.01506e-10 4.03583e-10     4.6e-09 4.59898e-09  9.5982e-09
"""
import logging
import time

import _matscipy
import numpy as np

import scipy.optimize
import scipy.spatial.distance


# https://stackoverflow.com/questions/21377020/python-how-to-do-lazy-debug-logging
class DeferredMessage(object):
    """Lazy evaluation for log messages."""

    def __init__(self, msg, func, *args, **kwargs):
        self.msg = msg
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.msg.format(self.func(*self.args, **self.kwargs))


def brute_force_closest_pair(x):
    """Find coordinate pair with minimum distance squared ||xi-xj||^2.

    Parameters
    ----------
    x: (N,dim) ndarray
      coordinates

    Returns
    -------
    float, (ndarray, ndarray): minimum distance squared and coodinates pair

    Examples
    --------
    Compare the performance of closest pair algorithms:

        >>> from matscipy.electrochemistry.steric_distribution import scipy_distance_based_target_function
        >>> from matscipy.electrochemistry.steric_distribution import numpy_only_target_function
        >>> from matscipy.electrochemistry.steric_distribution import brute_force_target_function
        >>> import itertools
        >>> import pandas as pd
        >>> import scipy.spatial.distance
        >>> import timeit
        >>>
        >>> funcs = [
        >>>         brute_force_closest_pair,
        >>>         scipy_distance_based_closest_pair,
        >>>         planar_closest_pair ]
        >>> func_names = ['brute','scipy','planar']
        >>> stats = []
        >>> N = 1000
        >>> dim = 3
        >>> for k in range(5):
        >>>     x = np.random.rand(N,dim)
        >>>     lambdas = [ (lambda x=x,f=f: f(x)) for f in funcs ]
        >>>     rets    = [ f() for f in lambdas ]
        >>>     vals    = [ v[0] for v in rets ]
        >>>     coords  = [ c for v in rets for p in v[1] for c in p ]
        >>>     times   = [ timeit.timeit(f,number=1) for f in lambdas ]
        >>>     diffs   = scipy.spatial.distance.pdist(
        >>>         np.atleast_2d(vals).T,metric='euclidean')
        >>>     stats.append((*vals,*diffs,*times,*coords))
        >>>
        >>> func_name_tuples = list(itertools.combinations(func_names,2))
        >>> diff_names =  [ 'd_{:s}_{:s}'.format(f1,f2) for (f1,f2) in func_name_tuples ]
        >>> perf_names =  [ 't_{:s}'.format(f) for f in func_names ]
        >>> coord_names = [
        >>>     'p{:d}{:s}_{:s}'.format(i,a,f) for f in func_names for i in (1,2) for a in ('x','y','z') ]
        >>> float_fields = [*func_names,*diff_names,*perf_names,*coord_names]
        >>> dtypes = [ (field, 'f4') for field in float_fields ]
        >>> labeled_stats = np.array(stats,dtype=dtypes)
        >>> stats_df = pd.DataFrame(labeled_stats)
        >>> print(stats_df.T.to_string(float_format='%8.6g'))
                                 0           1           2           3           4
        brute          2.24089e-05 5.61002e-05 8.51047e-05 3.48424e-05 5.37235e-05
        scipy          2.24089e-05 5.61002e-05 8.51047e-05 3.48424e-05 5.37235e-05
        planar         2.24089e-05 5.61002e-05 8.51047e-05 3.48424e-05 5.37235e-05
        d_brute_scipy            0           0           0           0           0
        d_brute_planar           0           0           0           0           0
        d_scipy_planar           0           0           0           0           0
        t_brute            4.02697     3.85543      4.1414     3.90338     3.86993
        t_scipy         0.00708364  0.00698962  0.00762594  0.00703242  0.00703579
        t_planar           0.38302     0.39462    0.434342    0.407233    0.420773
        p1x_brute         0.132014    0.331441    0.553405    0.534633    0.977582
        p1y_brute         0.599688    0.186959     0.90897    0.575864    0.636278
        p1z_brute          0.49631    0.993856    0.246418    0.853567    0.411793
        p2x_brute         0.134631    0.333526     0.55322    0.534493    0.977561
        p2y_brute         0.603598    0.179771    0.915063    0.576894    0.629313
        p2z_brute         0.496833    0.994145    0.239493    0.859377    0.409509
        p1x_scipy         0.132014    0.331441    0.553405    0.534633    0.977582
        p1y_scipy         0.599688    0.186959     0.90897    0.575864    0.636278
        p1z_scipy          0.49631    0.993856    0.246418    0.853567    0.411793
        p2x_scipy         0.134631    0.333526     0.55322    0.534493    0.977561
        p2y_scipy         0.603598    0.179771    0.915063    0.576894    0.629313
        p2z_scipy         0.496833    0.994145    0.239493    0.859377    0.409509
        p1x_planar        0.132014    0.331441     0.55322    0.534633    0.977561
        p1y_planar        0.599688    0.186959    0.915063    0.575864    0.629313
        p1z_planar         0.49631    0.993856    0.239493    0.853567    0.409509
        p2x_planar        0.134631    0.333526    0.553405    0.534493    0.977582
        p2y_planar        0.603598    0.179771     0.90897    0.576894    0.636278
        p2z_planar        0.496833    0.994145    0.246418    0.859377    0.411793
    """
    logger = logging.getLogger(__name__)
    t0 = time.perf_counter()

    n = x.shape[0]
    imin = 0
    jmin = 1
    if n < 2:
        return (None, None), float('inf')

    dx = x[0,:] - x[1,:]
    dxsq = np.square(dx)
    mindsq = np.sum(dxsq)

    for i in np.arange(n):
        for j in np.arange(i+1, n):
            dx = x[i,:] - x[j,:]
            dxsq = np.square(dx)
            dxnormsq = np.sum(dxsq)
            if dxnormsq < mindsq:
                imin = i
                jmin = j
                mindsq = dxnormsq

    t1 = time.perf_counter()-t0
    logger.debug("""Found minimum distance squared {:10.5e} for pair
        ({:d},{:d}) with coodinates {} and {} within {:10.5e} s.""".format(
        mindsq, imin, jmin, x[imin,:], x[jmin,:], t1))
    return mindsq, (x[imin,:], x[jmin,:])


def recursive_closest_pair(x, y):
    """Find coordinate pair with minimum distance squared ||xi-xj||^2.

    Find coordinate pair with minimum distance squared ||xi-xj||^2
    with one point from x and the other point from y

    Parameters
    ----------
    x: (N,dim) ndarray
        coordinates

    Returns
    -------
    float, (ndarray, ndarray): minimum distance squared and coodinate pair
    """
    # t0 = time.perf_counter()

    n = x.shape[0]
    if n < 4:
        return brute_force_closest_pair(x)

    mid = n // 2
    xl = x[:mid]
    xr = x[mid:]

    xdivider = x[mid,0]
    yl = []
    yr = []

    m = y.shape[0]
    for j in np.arange(m):
        if y[j,0] <= xdivider:
            yl.append(y[j,:])
        else:
            yr.append(y[j,:])

    yl = np.array(yl)
    yr = np.array(yr)
    mindsql, (pil, pjl) = recursive_closest_pair(xl, yl)
    mindsqr, (pir, pjr) = recursive_closest_pair(xr, yr)

    mindsq, (pim, pjm) = (mindsql, (pil, pjl)) if mindsql < mindsqr else (
        mindsqr, (pir, pjr))

    # TODO: this latter part only valid for 2d problems,
    # see https://sites.cs.ucsb.edu/~suri/cs235/ClosestPair.pdf
    # some 3d implementation at
    # https://github.com/eyny/closest-pair-3d/blob/master/src/ballmanager.cpp
    close_y = np.array(
        [y[j,:] for j in np.arange(m) if (np.square(y[j,0]-xdivider) < mindsq)])

    close_n = close_y.shape[0]
    if close_n > 1:
        for i in np.arange(close_n-1):
            for j in np.arange(i+1, min(i+8, close_n)):
                dx = close_y[i,:] - close_y[j,:]
                dxsq = np.square(dx)
                dxnormsq = np.sum(dxsq)
                if dxnormsq < mindsq:
                    pim = close_y[i,:]
                    pjm = close_y[j,:]
                    mindsq = dxnormsq

    return mindsq, (pim, pjm)


def planar_closest_pair(x):
    """Find coordinate pair with minimum distance ||xi-xj||.

    ATTENTION: this implementation tackles the planar problem!

    Parameters
    ----------
    x: (N,dim) ndarray
      coordinates

    Returns
    -------
    float, (ndarray, ndarray): minimum distance squared and coodinates pair
    """
    logger = logging.getLogger(__name__)
    assert isinstance(x, np.ndarray), "np.ndarray expected for x"
    assert x.ndim == 2, "x is expected to be 2d array"

    t0 = time.perf_counter()

    I = np.argsort(x[:,0])
    J = np.argsort(x[:,-1])
    X = x[I,:]
    Y = x[J,:]
    mindsq, (pim, pjm) = recursive_closest_pair(X,Y)

    # mind = np.sqrt(mindsq)
    t1 = time.perf_counter()-t0
    logger.debug("""Found minimum distance squared {:10.5e} for pair with
        coodinates {} and {} within {:10.5e} s.""".format(mindsq,pim,pjm,t1))
    return mindsq, (pim, pjm)


def scipy_distance_based_closest_pair(x):
    """Find coordinate pair with minimum distance ||xi-xj||.

    Parameters
    ----------
    x: (N,dim) ndarray
      coordinates

    Returns
    -------
    float, (ndarray, ndarray): minimum distance squared and coodinates pair

    Examples
    --------
    Handling condensed distance matrix indices:

        >>> c = np.array([1, 2, 3, 4, 5, 6])
        >>> print(c)
        [1 2 3 4 5 6]

        >>> d = scipy.spatial.distance.squareform(c)
        >>> print(d)
        [[0 1 2 3]
         [1 0 4 5]
         [2 4 0 6]
         [3 5 6 0]]

        >>> I,J = np.tril_indices(d.shape[0],-1)
        >>> print(I,J)
        [1 2 2 3 3 3] [0 0 1 0 1 2]

        >>> I,J = np.triu_indices(d.shape[0],1)
        >>> print(I,J)
        [0 0 0 1 1 2] [1 2 3 2 3 3]

        >>> print(d[I])
        [1 2 4 3 5 6]
    """
    logger = logging.getLogger(__name__)
    t0 = time.perf_counter()

    n = x.shape[0]

    dxnormsq = scipy.spatial.distance.pdist(x, metric='sqeuclidean')

    ij = np.argmin(dxnormsq)
    mindsq = dxnormsq[ij]

    # I,J = np.tril_indices(n,-1)
    I,J = np.triu_indices(n,1)
    imin,jmin = (I[ij],J[ij])

    t1 = time.perf_counter()-t0
    logger.debug("""Found minimum distance squared {:10.5e} for pair
        ({:d},{:d}) with coodinates {} and {} within {:10.5e} s.""".format(
            mindsq,imin,jmin,x[imin,:],x[jmin,:],t1))
    return mindsq, (x[imin,:], x[jmin,:])


def brute_force_target_function(x, r=1.0, constraints=None):
    """Target function. Penalize dense packing for coordinates ||xi-xj||<ri+rj.

    Parameters
    ----------
    x: (N,dim) ndarray
        particle coordinates
    r: float or (N,) ndarray, optional (default=1.0)
        steric radii of particles

    Returns
    -------
    float: target function value

    Examples
    --------
    Compare performance of target functions:


        >>> from matscipy.electrochemistry.steric_distribution import scipy_distance_based_target_function
        >>> from matscipy.electrochemistry.steric_distribution import numpy_only_target_function
        >>> from matscipy.electrochemistry.steric_distribution import brute_force_target_function
        >>> import itertools
        >>> import pandas as pd
        >>> import scipy.spatial.distance
        >>> import timeit
        >>>
        >>> funcs = [
        >>>         brute_force_target_function,
        >>>         numpy_only_target_function,
        >>>         scipy_distance_based_target_function ]
        >>> func_names = ['brute','numpy','scipy']
        >>>
        >>> stats = []
        >>> K = np.exp(np.log(10)*np.arange(-3,3))
        >>> for k in K:
        >>>     lambdas = [ (lambda x0=x0,k=k,f=f: f(x0*k)) for f in funcs ]
        >>>     vals    = [ f() for f in lambdas ]
        >>>     times   = [ timeit.timeit(f,number=1) for f in lambdas ]
        >>>     diffs = scipy.spatial.distance.pdist(np.atleast_2d(vals).T,metric='euclidean')
        >>>     stats.append((k,*vals,*diffs,*times))
        >>>
        >>> func_name_tuples = list(itertools.combinations(func_names,2))
        >>> diff_names = [ 'd_{:s}_{:s}'.format(f1,f2) for (f1,f2) in func_name_tuples ]
        >>> perf_names = [ 't_{:s}'.format(f) for f in func_names ]
        >>> fields =  ['k',*func_names,*diff_names,*perf_names]
        >>> dtypes = [ (field, '>f4') for field in fields ]
        >>> labeled_stats = np.array(stats,dtype=dtypes)
        >>> stats_df = pd.DataFrame(labeled_stats)
        >>> print(stats_df.to_string(float_format='%8.6g'))
                 k       brute       numpy       scipy  d_brute_numpy  d_brute_scipy  d_numpy_scipy  t_brute  t_numpy   t_scipy
        0    0.001  3.1984e+07  3.1984e+07  3.1984e+07    5.58794e-08    6.70552e-08    1.11759e-08 0.212432 0.168858 0.0734278
        1     0.01 3.19829e+07 3.19829e+07 3.19829e+07    9.31323e-08    7.82311e-08    1.49012e-08 0.212263  0.16846 0.0791856
        2      0.1 3.18763e+07 3.18763e+07 3.18763e+07    7.45058e-09    1.86265e-08    1.11759e-08 0.201706 0.164867 0.0711544
        3        1 2.27418e+07 2.27418e+07 2.27418e+07    3.72529e-08    4.84288e-08    1.11759e-08  0.20762 0.166005 0.0724238
        4       10      199751      199751      199751    1.16415e-10    2.91038e-11    8.73115e-11 0.202635 0.161932 0.0772684
        5      100     252.548     252.548     252.548    3.28555e-11              0    3.28555e-11 0.202512 0.161217 0.0726705
    """
    logger = logging.getLogger(__name__)
    assert x.ndim == 2, "2d array expected for x"
    # sum_i sum_{j!=i} max(0,(r_i+r_j)"^2-||xi-xj||^2)^2
    f = 0
    n = x.shape[0]
    xi = x

    ri = r
    if not isinstance(r, np.ndarray) or r.shape != (n,):
        ri = ri*np.ones(n)
    assert  ri.shape == (n,)

    zeros = np.zeros(n)
    for i in np.arange(1,n):
        rj = np.roll(ri, i, axis=0)
        xj = np.roll(xi, i, axis=0)
        d = ri + rj
        dsq = np.square(d)
        dx = xi - xj
        dxsq = np.square(dx)
        dxnormsq = np.sum(dxsq, axis=1)
        sqdiff = dsq - dxnormsq
        penalty = np.maximum(zeros, sqdiff)
        penaltysq = np.square(penalty)
        # half for double-counting
        f += 0.5*np.sum(penaltysq)

    if constraints:
        logger.debug(
            "Unconstrained penalty: {:10.5e}.".format(f))
        f += constraints(x)

    logger.debug(
        "Total penalty:         {:10.5e}.".format(f))
    return f


# TODO: code explicit Jacobian
def scipy_distance_based_target_function(x, r=1.0, constraints=None):
    """Target function. Penalize dense packing for coordinates ||xi-xj||<ri+rj.

    Parameters
    ----------
    x: (N,dim) ndarray
        particle coordinates
    r: float or (N,) ndarray, optional (default=1.0)
        steric radii of particles

    Returns
    -------
    float: target function value
    """
    logger = logging.getLogger(__name__)
    assert x.ndim == 2, "2d array expected for x"
    # sum_i sum_{j!=i} max(0,(r_i+r_j)"^2-||xi-xj||^2)^2
    n = x.shape[0]

    if not isinstance(r, np.ndarray) or r.shape != (n,):
        r = r*np.ones(n)
    assert r.shape == (n,)

    # r(Nx1) kron ones(1xN) = Ri(NxN)
    Ri = np.kron(r, np.ones((n, 1)))
    Rj = Ri.T
    Dij = Ri + Rj
    dij = scipy.spatial.distance.squareform(
        Dij, force='tovector', checks=False)

    zeros = np.zeros(dij.shape)

    dsq = np.square(dij)

    dxnormsq = scipy.spatial.distance.pdist(x, metric='sqeuclidean')
    # computes the squared Euclidean distance ||u-v||_2^2 between vectors
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html

    sqdiff = dsq - dxnormsq

    penalty = np.maximum(zeros, sqdiff)
    penaltysq = np.square(penalty)
    # no double-counting here
    f = np.sum(penaltysq)

    if constraints:
        logger.debug(
            "Unconstrained penalty: {:10.5e}.".format(f))
        f += constraints(x)

    logger.debug(
            "Total penalty:         {:10.5e}.".format(f))
    return f


# https://www.researchgate.net/publication/266617010_NumPy_SciPy_Recipes_for_Data_Science_Squared_Euclidean_Distance_Matrices
def numpy_only_target_function(x, r=1.0, constraints=None):
    """Target function. Penalize dense packing for coordinates ||xi-xj||<ri+rj.

    Parameters
    ----------
    x: (N,dim) ndarray
        particle coordinates
    r: float or (N,) ndarray, optional (default=1.0)
        steric radii of particles

    Returns
    -------
    float: target function value
    """
    logger = logging.getLogger(__name__)
    assert x.ndim == 2, "2d array expected for x"
    # sum_i sum_{j!=i} max(0,(r_i+r_j)"^2-||xi-xj||^2)^2
    n = x.shape[0]

    if not isinstance(r, np.ndarray) or r.shape != (n,):
        r = r*np.ones(n)
    assert r.shape == (n,)

    zeros = np.zeros((n,n))

    # r(Nx1) kron  ones(1xN) = Ri(NxN)
    Ri = np.kron(r, np.ones((n,1)))
    Rj = Ri.T
    Dij = Ri + Rj
    dsq = np.square(Dij)
    np.fill_diagonal(dsq,0.)

    G = np.dot(x,x.T)
    H = np.tile(np.diag(G), (n,1))
    dxnormsq = H + H.T - 2*G
    sqdiff = dsq - dxnormsq

    penalty = np.maximum(zeros, sqdiff)
    penaltysq = np.square(penalty)
    # half for double-counting
    f = 0.5*np.sum(penaltysq)

    if constraints:
        logger.debug(
            "Unconstrained penalty: {:10.5e}.".format(f))
        f += constraints(x)

    logger.debug(
        "Total penalty:         {:10.5e}.".format(f))
    return f


def neigh_list_based_target_function(x, r=1.0, constraints=None, Dij=None):
    """Target function. Penalize dense packing for coordinates ||xi-xj||<ri+rj.

    Parameters
    ----------
    x: (N,dim) ndarray
        particle coordinates
    r: float or (N,) ndarray, optional (default=1.0)
        steric radii of particles
    constraints: callable, returns (float, (N,dim) ndarray)
        constraint function value and gradien
    Dij: (N, N) ndarray, optional (default=None)
        pairwise minimum allowed distance matrix, overrides r

    Returns
    -------
    (float, (N,dim) ndarray): target function value and gradient

        f: float, target (or penalty) function, value evaluates to
              sum_i sum_{j!=i} max(0,(r_i+r_j)^2-||xi-xj||^2)^2
            without double-counting pairs, where r_i and r_j are
            the steric radii of coordinate points i and j
        df: (N,dim) ndarray of float, the gradient, evaluates to
              4*si sj ((r_i'+r_j)^2-||xi-xj||^2)^2)*(xik'-xjk')*(kdi'j-kdi'i)
            for entry (i',k'), where i subscribes the coordinate point and k
            the spatial dimension. kd is Kronecker delta, si is sum over i
    """
    logger = logging.getLogger(__name__)
    #
    # function:
    #
    # f(x) = sum_i sum_{j!=i} max(0,(r_i+r_j)^2-||xi-xj||^2)^2
    #
    # gradient:
    #
    # dfdxi'k'=4*si sj ((r_i'+r_j)^2-||xi-xj||^2)^2)*(xik'-xjk')*(kdi'j-kdi'i)
    #
    # for all pairs i'j within ri'+rj cutoff
    # where k is the spatial direction x, y or z.

    assert x.ndim == 2, "2d array expected for x"
    n = x.shape[0]

    if not Dij:
        if not isinstance(r, np.ndarray) or r.shape != (n,):
            r = r*np.ones(n)
        assert r.shape == (n,)

        # compute minimum allowed pairwise distances Dij (NxN matrix)
        # r(Nx1) kron ones(1xN) = Ri(NxN)
        Ri = np.kron(r, np.ones((n, 1)))
        Rj = Ri.T
        Dij = Ri + Rj

    # TODO: allow for periodic boundaries, for now use shrink wrapped box
    box = np.array([x.min(axis=0), x.max(axis=0)])
    cell_origin = box[0,:]
    cell = np.diag(box[1,:] - box[0,:])

    # get all pairs within their allowed minimum distance
    # parameters are
    # _matscipy.neighbour_list(quantities, cell_origin, cell,
    #                          np.linalg.inv(cell.T), pbc, positions,
    #                          cutoff, numbers)
    # If thhe parameter 'cutoff' is a per-atom value, then it must be a
    # diamater, not a radius (as wrongly stated within the function's
    # docstring)
    i, j, dxijnorm, dxijvec = _matscipy.neighbour_list(
        'ijdD', cell_origin, cell, np.linalg.inv(cell.T), [0,0,0],
        x, 2.0*Ri, np.ones(len(x),dtype=np.int32))
    # i, j are coordinate point indices, dxijnorm is pairwise distance,
    #   dxvijvec is distance vector

    # nl contains redundnancies, i.e. ij AND ji
    # pairs = list(zip(i,j))
    logger.debug("Number of pairs within minimum allowed distance: {:d}"
                 .format(len(i)))

    # get minimum allowed pairwise distance for all pairs within this distance
    dij = Dij[i,j]

    # (r_i'+r_j)^2
    dsq = np.square(dij)

    # ||xi-xj||^2
    dxnormsq = np.square(dxijnorm)

    # (r_i+r_j)^2-||xi-xj||^2
    sqdiff = dsq - dxnormsq

    # ((r_i+r_j)^2-||xi-xj||^2)^2
    penaltysq = np.square(sqdiff)

    # correct for double counting due to redundancies
    f = 0.5*np.sum(penaltysq)

    # 4*si sj ((r_i'+r_j)^2-||xi-xj||^2)^2)*(xik'-xjk')*(kdi'j-kdi'i)
    # (N x 1) column vector (a1 ... aN) * (N x dim) matrix ((d11,))
    # Other than the formula above implicates, _matscipy.neighbour_list
    # returns the distance vector dxijvec = xj-xi (pointing from i to j),
    # thus positive sign in gradient below:
    gradij = 4*np.atleast_2d(sqdiff).T*dxijvec  # let's hope for proper broadcasting
    grad = np.zeros(x.shape)
    grad[i] += gradij
    # Since neighbour list always includes ij and ji, we only have to treat i,
    # not j. The following line is obsolete (otherwise we would introduce
    # double-counting again):
    # grad[j] -= gradij

    if constraints:
        logger.debug(
            "Unconstrained penalty:       {:10.5e}.".format(f))
        logger.debug(DeferredMessage(
            "Unconstrained gradient norm: {:10.5e}.",
            np.linalg.norm, grad))
        g, g_grad = constraints(x)
        f += g
        grad += g_grad

    grad_1d = grad.reshape(np.product(grad.shape))

    logger.debug(
        "Total penalty:               {:10.5e}.".format(f))
    logger.debug(DeferredMessage(
        "Gradient norm:               {:10.5e}.",
        np.linalg.norm, grad_1d))

    return f, grad_1d


def box_constraint(x, box=np.array([[0., 0., 0], [1.0, 1.0, 1.0]]), r=0.):
    """Constraint function. Confine coordinates within box.

    Parameters
    ----------
    x: (N,dim) ndarray
      coordinates
    box: (2,dim) ndarray, optional (default: [[0.,0.,0.],[1.,1.,1.]])
      box corner coordinates
    r: float or np.(N,) ndarray, optional (default=0)
      steric radii of particles

    Returns
    -------
    float: penalty, positive
    """
    logger = logging.getLogger(__name__)
    zeros = np.zeros(x.shape)
    r = np.atleast_2d(r).T

    # positive if coordinates out of box
    ldist = box[0, :] - x + r
    rdist = x + r - box[1, :]

    lpenalty = np.maximum(zeros, ldist)
    rpenalty = np.maximum(zeros, rdist)

    lpenaltysq = np.square(lpenalty)
    rpenaltysq = np.square(rpenalty)

    g = np.sum(lpenaltysq) + np.sum(rpenaltysq)
    logger.debug("Constraint penalty: {:.4g}.".format(g))
    return g


def box_constraint_with_gradient(
        x, box=np.array([[0., 0., 0], [1.0, 1.0, 1.0]]), r=0.):
    """Constraint function. Confine coordinates within box.

    Parameters
    ----------
    x: (N,dim) ndarray
      coordinates
    box: (2,dim) ndarray, optional (default: [[0.,0.,0.],[1.,1.,1.]])
      box corner coordinates
    r: float or np.(N,) ndarray, optional (default=0)
      steric radii of particles

    Returns
    -------
    (float, (N,dim) ndarray of float): penalty g(x) and its gradient dg(x)

        g(x), positive:
                lik = box[0,k] - xik + ri  > 0 for xik out of lower boundaries
                rik = xik + ri - box[1,k]  > 0 for xik out of upper boundaries
            where box[0] and box[1] are lower and upper corner of bounding box
            and k marks spatial dimension.
                g(x) = sum_i sum_k ( max(0,lik) + max(0,rik) )^2

        dg(x): gradient, entries accordingly evaluates to
                dgdxik(x) = 2*( -max(0,lik) + max(0,rik) )
    """
    logger = logging.getLogger(__name__)
    zeros = np.zeros(x.shape)
    r = np.atleast_2d(r).T

    # positive if coordinates out of box
    ldist = box[0, :] - x + r
    rdist = x + r - box[1, :]

    lpenalty = np.maximum(zeros, ldist)
    rpenalty = np.maximum(zeros, rdist)

    lpenaltysq = np.square(lpenalty)
    rpenaltysq = np.square(rpenalty)

    g = np.sum(lpenaltysq) + np.sum(rpenaltysq)
    logger.debug("Constraint penalty: {:.4g}.".format(g))

    grad = -2*lpenalty + 2*rpenalty
    logger.debug(DeferredMessage(
        "Norm of constraint penalty gradient: {:.4g}.",
        np.linalg.norm, grad))

    return g, grad


def apply_steric_correction(
        x, box=None, r=None,
        method='L-BFGS-B',
        options={'gtol':1.e-8,'maxiter':100,'disp':True,'eps':1.0e-8},
        target_function=neigh_list_based_target_function,
        returns_gradient=True,
        closest_pair_function=scipy_distance_based_closest_pair):
    """Enforce steric constraints on coordinate distribution within box.

    Parameters
    ----------
    x : (N,dim) ndarray
        Particle coordinates.
    box: (2,dim) ndarray, optional (default: None)
        Box corner coordinates.
    r : float or (N,) ndarray, optional (default=None)
        Steric radius of particles. Can be specified particle-wise.
    options : dict, optional
        Forwarded to scipy minimzer.
        https://docs.scipy.org/doc/scipy/reference/optimize.minimize-bfgs.html
        (default: {'gtol':1.e-5,'maxiter':10,'disp':True,'eps':1.e-8})
    target_function: func, optional
        One of the target functions within this submodule, or function
        of same signature. (default: neigh_list_based_target_function)
    returns_gradient: bool, optional (default: Trze)
        If True, then 'target_function' is expected to return a tuple (f, df)
        with f the actual target function value and df its (N,dim) gradient.
        This flag must be set for 'neigh_list_based_target_function'.
    closest_pair_function: func, optional
        One of the closest pair functions within this submodule, or function
        of same signature. (default: scipy_distance_based_closest_pair)

    Returns
    -------
    float : (N,dim) ndarray
        Modified particle coordinates, meeting steric constraints.
    """
    logger = logging.getLogger(__name__)
    assert isinstance(x, np.ndarray), "x must be np.ndarray"
    assert x.ndim == 2, "x must be 2d array"

    n = x.shape[0]
    dim = x.shape[1]

    if r is None:
        r = 0.0
        logger.info("No steric radii explicitly specified, using none (zero).")

    r = np.atleast_1d(r)

    assert r.ndim == 1, "only isotropic steric radii r, no spatial dimensions"
    if r.shape[0] == 1:
        r = r*np.ones(n)
    assert r.shape[0] == n, "either one steric radius for all paricles or one each"

    if box is None:
        box = np.array(x.min(axis=0), x.max(axis=0))
        logger.info("No bounding box explicitly specified, using extreme")
        logger.info("coordinates ({}) of coordinate set as default.".format(
            box))

    assert isinstance(box, np.ndarray), "box must be np.ndarray"
    assert x.ndim == 2, "box must be 2d array"
    assert box.shape[0] == 2, "box must have two rows for outer corners"
    assert box.shape[1] == dim, "spatial dimensions of x and box must agree"

    V = np.product(box[1, :]-box[0, :])
    L = np.power(V, (1./dim))
    logger.info("Normalizing coordinates by reference length")
    logger.info("    L = V^(1/dim) = ({:.2g})^(1/{:d}) = {:.2g}.".format(
        V, dim, L))

    # normalizing to unit volume necessary,
    # as target function apparently not dimension-insensitive
    BOX = box / L
    X0 = x / L
    R = r / L

    logger.info("Normalized bounding box: ")
    logger.info("    {}.".format(BOX[0]))
    logger.info("    {}.".format(BOX[1]))

    # flatten coordinates for scipy optimizer
    x0 = X0.reshape(np.product(X0.shape))


    # define constraint and target wrapper for scipy optimizer
    if returns_gradient:

        def g(x):
            return box_constraint_with_gradient(x, box=BOX, r=R)

        def f(x):
            f, grad = target_function(x.reshape((n, dim)), r=R, constraints=g)
            return f, grad.reshape(np.product(grad.shape))

        gval, ggrad = g(X0)
        fval, fgrad = f(x0)
        logger.info("Initial constraint penalty:       {:10.5e}.".format(gval))
        logger.info("Initial total penalty:            {:10.5e}.".format(fval))
        logger.info("Initial constraint gradient norm: {:10.5e}.".format(
            np.linalg.norm(ggrad)))
        logger.info("Initial total gradient norm:      {:10.5e}.".format(
            np.linalg.norm(fgrad)))

    else:

        def g(x):
            return box_constraint(x, box=BOX, r=R)

        def f(x):
            return target_function(x.reshape((n, dim)), r=R, constraints=g)

        logger.info("Initial constraint penalty: {:10.5e}.".format(g(X0)))
        logger.info("Initial total penalty:      {:10.5e}.".format(f(x0)))

    # log initial minimum distance between pairs
    minDsq, (P1, P2) = closest_pair_function(X0)  # dimensionless
    mindsq, (p1, p2) = closest_pair_function(x)  # dimensional

    minD = np.sqrt(minDsq)
    mind = np.sqrt(mindsq)

    # logger.info("Final distribution has residual penalty {:10.5e}.".format(f1))
    logger.info("Min. dist. {:10.5e} between points".format(mind))
    logger.info("    {} and".format(p1))
    logger.info("    {}.".format(p2))
    logger.info("Min. dist. {:10.5e} between dimensionless points".format(minD))
    logger.info("    {} and".format(P1))
    logger.info("    {}.".format(P2))
    logger.info("    normalized by L = {:.2g}.".format(L))




    def minimizer_callback(xk, *_):
        """Callback function to be used by optimizers of scipy.optimize.

        The second argument "*_" makes sure that it still works when the
        optimizer calls the callback function with more than one argument. See
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
        """
        nonlocal closest_pair_function, callback_count, tk, returns_gradient
        if callback_count == 0 and returns_gradient:
            logger.info(
                "{:>12s} {:>12s} {:>12s} {:>12s} {:>12s} {:>12s}".format(
                    "#callback", "objective", "gradient", "min. dist.",
                    "timing, step", "timing, tot."))
        elif callback_count == 0:
            logger.info(
                "{:>12s} {:>12s} {:>12s} {:>12s} {:>12s}".format(
                    "#callback", "objective", "min. dist.",
                    "timing, step", "timing, tot."))

        if returns_gradient:
            fk, gradk = f(xk)
            normgradk = np.linalg.norm(gradk)
        else:
            fk = f(xk)

        Xk = xk.reshape((n, dim))
        mindsq, _ = closest_pair_function(Xk)
        mind = np.sqrt(mindsq)

        t1 = time.perf_counter()
        dt = t1 - tk
        dT = t1 - t0
        tk = t1

        if returns_gradient:
            logger.info(
                "{:12d} {:12.5e} {:12.5e} {:12.5e} {:12.5e} {:12.5e}".format(
                    callback_count, fk, normgradk, mind, dt, dT))
        else:
            logger.info(
                "{:12d} {:12.5e} {:12.5e} {:12.5e} {:12.5e}".format(
                    callback_count, fk, mind, dt, dT))

        callback_count += 1

    callback_count = 0
    t0 = time.perf_counter()
    tk = t0  # previosu callback timer value
    # call once for initial configuration
    if logger.isEnabledFor(logging.INFO):
        callback = minimizer_callback
        callback(x0)
    else:
        callback = None

    # neat lecture on scipy optimizers
    # http://scipy-lectures.org/advanced/mathematical_optimization/
    res = scipy.optimize.minimize(f, x0, method=method, jac=returns_gradient,
                                  callback=callback, options=options)

    if not res.success:
        logger.warn(res.message)

    x1 = res.x  # dimensionless, flat
    X1 = x1.reshape((n, dim))  # dimensionless, 2d

    if returns_gradient:
        gval, ggrad = g(X1)
        fval, fgrad = f(x1)
        logger.info("Final constraint penalty:       {:10.5e}.".format(gval))
        logger.info("Final total penalty:            {:10.5e}.".format(fval))
        logger.info("Final constraint gradient norm: {:10.5e}.".format(
            np.linalg.norm(ggrad)))
        logger.info("Final total gradient norm:      {:10.5e}.".format(
            np.linalg.norm(fgrad)))

    else:
        logger.info("Final constraint penalty: {:10.5e}.".format(g(X1)))
        logger.info("Final total penalty:      {:10.5e}.".format(f(x1)))

    x1 = X1*L  # dimensional
    minDsq, (P1, P2) = closest_pair_function(X1)  # dimensionless
    mindsq, (p1, p2) = closest_pair_function(x1)  # dimensional

    minD = np.sqrt(minDsq)
    mind = np.sqrt(mindsq)

    # logger.info("Final distribution has residual penalty {:10.5e}.".format(f1))
    logger.info("Min. dist. {:10.5e} between points".format(mind))
    logger.info("    {} and".format(p1))
    logger.info("    {}.".format(p2))
    logger.info("Min. dist. {:10.5e} between dimensionless points".format(minD))
    logger.info("    {} and".format(P1))
    logger.info("    {}.".format(P2))
    logger.info("    normalized by L = {:.2g}.".format(L))

    dT = time.perf_counter() - t0
    logger.info("Ellapsed time: {:10.5} s.".format(dT))

    return x1, res



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""Generate discrete coordinate sets from continuous distributions.
Export as atom positions in .xyz of LAMMPS data file format.

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:
  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
  Lukas Elflein <elfleinl@cs.uni-freiburg.de>
"""
import argparse
import logging
import os
import os.path
import sys

import numpy as np

import ase
import ase.io
import scipy.constants as sc
from scipy import interpolate, integrate

from matscipy.electrochemistry import continuous2discrete

logger = logging.getLogger(__name__)


def main():
    """Generate discrete coordinate sets from continuous distributions.
    Export as atom positions in .xyz of LAMMPS data file format.
    Plot continuous and discrete distributions if wanted.

    ATTENTION: LAMMPS data file export (atom style 'full') requires
    ase>=3.19.0b1 (> 6th Nov 2019) due to recently reseolved issue"""

    # in order to have both:
    # * preformatted help text and ...
    # * automatic display of defaults
    class ArgumentDefaultsAndRawDescriptionHelpFormatter(
            argparse.ArgumentDefaultsHelpFormatter,
            argparse.RawDescriptionHelpFormatter):
        pass

    class StoreAsNumpyArray(argparse._StoreAction):
        def __call__(self, parser, namespace, values, option_string=None):
            values = np.array(values, ndmin=1)
            return super().__call__(parser, namespace, values, option_string)

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)

    parser.add_argument('infile', metavar='IN', nargs='?',
                        help='binary numpy .npz or plain text .txt input file')
    parser.add_argument('outfile', metavar='OUT', nargs='?',
                        help='.xyz format output file')

    parser.add_argument('--box', '-b', default=[50.0e-9, 50.0e-9, 100.0e-9], nargs=3,
                        action=StoreAsNumpyArray,
                        metavar=('X', 'Y', 'Z'), required=False, type=float,
                        dest="box", help='Box dimensions')

    parser.add_argument('--names', default=['Na', 'Cl'], type=str, nargs='+',
                        metavar=('NAME'), required=False, dest="names",
                        help='Atom names')

    parser.add_argument('--charges', default=[1, -1], type=float, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('NAME'), required=False, dest="charges",
                        help='Atom charges')

    parser.add_argument('--mol-id-offset', default=[-1, -1],
                        type=int, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('OFFSET'), required=False,
                        dest="mol_id_offset",
                        help=('When storing as LAMMPS data, this tool uses'
                              ' atom sytle "full", assigning a molecule id to'
                              ' each atom. Per default, a unique 1-indexed'
                              ' molecule id corrsponding to its atom id is'
                              ' assigned to each atom. Specify a species-wise'
                              ' offset here. Specifying 0 (zero) for each'
                              ' species will result in moleucules containing '
                              ' one atom per species.'))

    # sampling
    parser.add_argument('--ngridpoints', default=np.nan, type=float, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('N'), required=False, dest="n_gridpoints",
                        help=('Number of grid points for discrete support. '
                              'Continuous support for all sampes per default. '
                              'Specify "NaN" explicitly for continuous support '
                              'in particular species, i.e. '
                              '"--n_gridpoints 100 NaN 50"'))
    parser.add_argument('--sample-size', default=np.nan, type=float, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('N'), required=False, dest="sample_size",
                        help=('Sample size. Specify '
                              'multiple values for specific number of atom '
                              'positions for each species. Per default, infer '
                              'sample size from distributions, assuming '
                              'concentrations in SI units (i.e. mM or mol / m^3).'
                              'Specify "NaN" explicitly for inference in certain '
                              'species only, i.e. '
                              '"--sample-size 100 NaN 50"'))

    # output
    parser.add_argument('--nbins', default=100, type=int,
                        metavar=('N'), required=False, dest="nbins",
                        help='Number of bins for histogram plots')

    parser.add_argument('--debug', default=False, required=False,
                        action='store_true', dest="debug", help='debug flag')
    parser.add_argument('--verbose', default=False, required=False,
                        action='store_true', dest="verbose", help='verbose flag')
    parser.add_argument('--log', required=False, nargs='?', dest="log",
                        default=None, const='c2d.log', metavar='LOG',
                        help='Write log file c2d.log, optionally specify log file name')

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
        # This supports bash autocompletion. To enable this, pip install
        # argcomplete, activate global completion, or add
        #      eval "$(register-python-argcomplete lpad)"
        # into your .bash_profile or .bashrc
    except ImportError:
        pass

    args = parser.parse_args()

    if args.debug:
        loglevel = logging.DEBUG
    elif args.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    logformat = "[ %(filename)s:%(lineno)s - %(funcName)s() ]: %(message)s"

    logging.basicConfig(level=loglevel,
                        format=logformat)

    # explicitly modify the root logger (necessary?)
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # remove all handlers
    for h in logger.handlers:
        logger.removeHandler(h)

    # create and append custom handles
    ch = logging.StreamHandler()
    formatter = logging.Formatter(logformat)
    ch.setFormatter(formatter)
    ch.setLevel(loglevel)
    logger.addHandler(ch)

    if args.log:
        fh = logging.FileHandler(args.log)
        fh.setFormatter(formatter)
        fh.setLevel(loglevel)
        logger.addHandler(fh)

    logger.info('This is `{}` : `{}`.'.format(__file__, __name__))

    if not isinstance(args.box, np.ndarray):
        args.box = np.array(args.box, ndmin=1)
    if not isinstance(args.sample_size, np.ndarray):
        args.sample_size = np.array(args.sample_size, ndmin=1)
    if not isinstance(args.n_gridpoints, np.ndarray):
        args.n_gridpoints = np.array(args.n_gridpoints, ndmin=1)

    if not args.infile:
        infile = sys.stdin
        infile_format = '.txt'
    else:
        infile = args.infile
        _, infile_format = os.path.splitext(infile)

    if infile_format == '.npz':
        file = np.load(infile)
        x = file['x']
        u = file['u']
        c = file['c']
    else:  # elif infile_format == 'txt'
        data = np.loadtxt(infile, unpack=True)
        x = data[0, :]
        u = data[1, :]
        c = data[2:, :]

    if c.ndim > 1:
        C = [c[k, :] for k in range(c.shape[0])]
    else:
        C = [c]

    del c

    logger.info('Read {:d} concentration distributions.'.format(len(C)))
    sample_size = args.sample_size
    sample_size = (sample_size.repeat(len(C)) if sample_size.shape == (1,)
                   else sample_size)

    # distribution functions from concentrations;
    D = [interpolate.interp1d(x, c) for c in C]

    # infer sample size from integral over concentration distribution if
    # no explicit sample size given
    # TODO: over-estimates sample size when distribution highly nonlinear
    for i, s in enumerate(sample_size):
        if np.isnan(s):
            # average concentration in distribution over interval
            cave, _ = integrate.quad(D[i], 0, args.box[-1]) / args.box[-1]  # z direction
            # [V] = m^3, [c] = mol / m^3, [N_A] = 1 / mol
            sample_size[i] = int(np.round(args.box.prod()*cave * sc.Avogadro))
            logger.info('Inferred {} samples on interval [{},{}] m'.format(
                sample_size[i], 0, args.box[-1]))
            logger.info('for average concentration {} mM.'.format(cave))

    n_gridpoints = args.n_gridpoints  # assume n_gridpoints is np.ndarray
    n_gridpoints = n_gridpoints.repeat(len(C)) if n_gridpoints.shape == (1,) else n_gridpoints

    logger.info('Generating {} positions on {} support for species {}.'.format(
        sample_size, n_gridpoints, args.names))

    logger.info('Generating structure from distribution ...')
    struc = [continuous2discrete(
                distribution=d,
                box=args.box, count=sample_size[k],
                n_gridpoints=n_gridpoints[k]) for k, d in enumerate(D)]

    logger.info('Generated {:d} coordinate sets.'.format(len(struc)))

    logger.info('Creating ase.Atom objects ...')
    system = ase.Atoms(
        cell=args.box/sc.angstrom,
        pbc=[1, 1, 0])

    # 1-indexed LAMMPS data molecule id within ASE's io module.
    # We assume here that 'mol-id' is 0-indexed internally but converted to a
    system.new_array('mol-id', [], dtype=int)

    for i, s in enumerate(struc):
        logger.info('{:d} samples in coordinate set {:d}.'.format(len(s), i))
        new_species = ase.Atoms(
            symbols=args.names[i]*int(sample_size[i]),
            charges=[args.charges[i]]*int(sample_size[i]),
            positions=s/sc.angstrom)

        # per default, consecutive numbering, otherwise custom offest:
        mol_id_offset = len(system) \
            if args.mol_id_offset[i] < 0 else args.mol_id_offset[i]

        # without offset, LAMMPS molecule ids are 1-indexed
        # mol-id 0 is reserved for 'non-bonded' atoms
        new_species.new_array(
            'mol-id', 1 + mol_id_offset + np.arange(
                start=0, stop=len(new_species), step=1, dtype=int), dtype=int)

        system += new_species

    logger.info('Writing output file ...')

    if not args.outfile:
        outfile = sys.stdout
        outfile_format = '.xyz'
    else:
        outfile = args.outfile
        _, outfile_format = os.path.splitext(outfile)

    logger.info('Output format {} to {}.'.format(outfile_format, outfile))

    if outfile_format == '.lammps':
        ase.io.write(
            outfile, system,
            format='lammps-data',
            units="real",
            atom_style='full',
            specorder=args.names)
    # specorder shoudl make sure ASE assigns types in the same order as
    # species names have been specified on command line.

    else:  # elif outfile_format == '.xyz'
        ase.io.write(outfile, system, format='extxyz')

    logger.info('Done.')


if __name__ == '__main__':
    # Execute everything else
    main()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
""" Compute ion concentrations consistent with general
Poisson-Nernst-Planck (PNP) equations via FEniCS.

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:
  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
"""
import numpy as np
import scipy.interpolate
import fenics as fn

from matscipy.electrochemistry.poisson_nernst_planck_solver import PoissonNernstPlanckSystem


class Boundary(fn.SubDomain):
    """Mark a point to be within the domain boundary for fenics."""
    # Boundary causes crash kernel if __init__ doe not call super().__init__()
    def __init__(self, x0=0, tol=1e-14):
        super().__init__()
        self.tol = tol
        self.x0 = x0

    def inside(self, x, on_boundary):
        """Mark a point to be within the domain boundary for fenics."""
        return on_boundary and fn.near(x[0], self.x0, self.tol)


class PoissonNernstPlanckSystemFEniCS(PoissonNernstPlanckSystem):
    """Describes and solves a 1D Poisson-Nernst-Planck system,
    using log concentrations internally"""

    @property
    def X(self):
        return self.mesh.coordinates().flatten()  # only 1D

    def solve(self):
        """Evoke FEniCS FEM solver.

        Returns
        -------
        uij : (Ni,) ndarray
            potential at Ni grid points
        nij : (M,Nij) ndarray
            concentrations of M species at Ni grid points
        lamj: (L,) ndarray
            value of L Lagrange multipliers
        """

        # weak form and FEM scheme:

        # in the weak form, u and v are the trial and test functions associated
        # with the Poisson part, p and q the trial and test functions associated
        # with the Nernst-Planck part. lam and mu are trial and test fuctions
        # associated to constraints introduced via Lagrange multipliers.
        # w is the whole set of trial functions [u,p,lam]
        # W is the space all w live in.
        rho = 0
        for i in range(self.M):
            rho += self.z[i]*self.p[i]

        source = - 0.5 * rho * self.v * fn.dx

        laplace = fn.dot(fn.grad(self.u), fn.grad(self.v))*fn.dx

        poisson = laplace + source

        nernst_planck = 0
        for i in range(self.M):
            nernst_planck += fn.dot(
                    - fn.grad(self.p[i]) - self.z[i]*self.p[i]*fn.grad(self.u),
                    fn.grad(self.q[i])
                )*fn.dx

        # constraints set up elsewhere
        F = poisson + nernst_planck + self.constraints

        fn.solve(F == 0, self.w, self.boundary_conditions)

        # store results:

        wij = np.array([self.w(x) for x in self.X]).T

        self.uij = wij[0, :]  # potential
        self.nij = wij[1:(self.M+1), :]  # concentrations
        self.lamj = wij[(self.M+1):]  # Lagrange multipliers

        return self.uij, self.nij, self.lamj

    def boundary_L(self, x, on_boundary):
        """Mark left boundary. Returns True if x on left boundary."""
        return on_boundary and fn.near(x[0], self.x0_scaled, self.bctol)

    def boundary_R(self, x, on_boundary):
        """Mark right boundary. Returns True if x on right boundary."""
        return on_boundary and fn.near(x[0], self.x1_scaled, self.bctol)

    def boundary_C(self, x, on_boundary):
        """Mark domain center."""
        return fn.near(x[0], (self.x0_scaled + self.x1_scaled)/2., self.bctol)

    def applyLeftPotentialDirichletBC(self, u0):
        """FEniCS Dirichlet BC u0 for potential at left boundary."""
        self.boundary_conditions.extend([
            fn.DirichletBC(self.W.sub(0), u0,
                           lambda x, on_boundary: self.boundary_L(x, on_boundary))])

    def applyRightPotentialDirichletBC(self, u0):
        """FEniCS Dirichlet BC u0 for potential at right boundary."""
        self.boundary_conditions.extend([
            fn.DirichletBC(self.W.sub(0), u0,
                           lambda x, on_boundary: self.boundary_R(x, on_boundary))])

    def applyLeftConcentrationDirichletBC(self, k, c0):
        """FEniCS Dirichlet BC c0 for k'th ion species at left boundary."""
        self.boundary_conditions.extend([
            fn.DirichletBC(self.W.sub(k+1), c0,
                           lambda x, on_boundary: self.boundary_L(x, on_boundary))])

    def applyRightConcentrationDirichletBC(self, k, c0):
        """FEniCS Dirichlet BC c0 for k'th ion species at right boundary."""
        self.boundary_conditions.extend([
            fn.DirichletBC(self.W.sub(k+1), c0,
                           lambda x, on_boundary: self.boundary_R(x, on_boundary))])

    def applyCentralReferenceConcentrationConstraint(self, k, c0):
        """FEniCS Dirichlet BC c0 for k'th ion species at right boundary."""
        self.boundary_conditions.extend([
            fn.DirichletBC(self.W.sub(k+1), c0,
                           lambda x, on_boundary: self.boundary_C(x, on_boundary))])

    # TODO: Robin BC!
    def applyLeftPotentialRobinBC(self, u0, lam0):
        self.logger.warning("Not implemented!")

    def applyRightPotentialRobinBC(self, u0, lam0):
        self.logger.warning("Not implemented!")

    def applyNumberConservationConstraint(self, k, c0):
        """
        Enforce number conservation constraint via Lagrange multiplier.
        See https://fenicsproject.org/docs/dolfin/1.6.0/python/demo/documented/neumann-poisson/python/documentation.html
        """
        self.constraints += self.lam[k]*self.q[k]*fn.dx \
            + (self.p[k]-c0)*self.mu[k]*fn.dx

    def applyPotentialDirichletBC(self, u0, u1):
        """Potential Dirichlet BC u0 and u1 on left and right boundary."""
        self.applyLeftPotentialDirichletBC(u0)
        self.applyRightPotentialDirichletBC(u1)

    def applyPotentialRobinBC(self, u0, u1, lam0, lam1):
        """Potential Robin BC on left and right boundary."""
        self.applyLeftPotentialRobinBC(u0, lam0)
        self.applyRightPotentialRobinBC(u1, lam1)

    def useStandardInterfaceBC(self):
        """Interface at left hand side and open bulk at right hand side"""
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled
        self.u1 = 0

        self.logger.info('Left hand side Dirichlet boundary condition:  u0 = {:> 8.4g}'.format(self.u0))
        self.logger.info('Right hand side Dirichlet boundary condition: u1 = {:> 8.4g}'.format(self.u1))

        self.applyPotentialDirichletBC(self.u0, self.u1)

        for k in range(self.M):
            self.logger.info(('Ion species {:02d} right hand side concentration '
                              'Dirichlet boundary condition: c1 = {:> 8.4g}').format(k, self.c_scaled[k]))
            self.applyRightConcentrationDirichletBC(k, self.c_scaled[k])

    def useStandardCellBC(self):
        """
        Interfaces at left hand side and right hand side, species-wise
        number conservation within interval."""
        self.boundary_conditions = []

        # Introduce a Lagrange multiplier per species anderson
        # rebuild discretization scheme (function spaces)
        self.K = self.M
        self.discretize()

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.
        self.u1 = - self.delta_u_scaled / 2.
        self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
          'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
        self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
          'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))

        self.applyPotentialDirichletBC(self.u0, self.u1)

        # Number conservation constraints
        self.constraints = 0
        N0 = self.L_scaled*self.c_scaled  # total amount of species in cell
        for k in range(self.M):
            self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
                'Ion species {:02d} number conservation constraint'.format(k),
                N0[k], lwidth=self.label_width))
            self.applyNumberConservationConstraint(k, self.c_scaled[k])

    def useCentralReferenceConcentrationBasedCellBC(self):
        """
        Interfaces at left hand side and right hand side, species-wise
        concentration fixed at cell center."""
        self.boundary_conditions = []

        # Introduce a Lagrange multiplier per species anderson
        # rebuild discretization scheme (function spaces)
        # self.K = self.M
        # self.discretize()

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.
        self.u1 = - self.delta_u_scaled / 2.
        self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
          'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
        self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
          'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))

        self.applyPotentialDirichletBC(self.u0, self.u1)

        for k in range(self.M):
            self.logger.info(
                'Ion species {:02d} reference concentration condition: c1 = {:> 8.4g} at cell center'.format(
                    k, self.c_scaled[k]))
            self.applyCentralReferenceConcentrationConstraint(k, self.c_scaled[k])

    def useSternLayerCellBC(self):
        """
        Interfaces at left hand side and right hand side, species-wise
        number conservation within interval."""
        self.boundary_conditions = []

        # Introduce a Lagrange multiplier per species anderson
        # rebuild discretization scheme (function spaces)
        self.constraints = 0
        self.K = self.M
        self.discretize()

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.
        self.u1 = - self.delta_u_scaled / 2.

        boundary_markers = fn.MeshFunction(
            'size_t', self.mesh, self.mesh.topology().dim()-1)

        bx = [
            Boundary(x0=self.x0_scaled, tol=self.bctol),
            Boundary(x0=self.x1_scaled, tol=self.bctol)]

        # Boundary.mark crashes the kernel if Boundary is internal class
        for i, b in enumerate(bx):
            b.mark(boundary_markers, i)

        boundary_conditions = {
            0: {'Robin': (1./self.lambda_S_scaled, self.u0)},
            1: {'Robin': (1./self.lambda_S_scaled, self.u1)},
        }

        ds = fn.Measure('ds', domain=self.mesh, subdomain_data=boundary_markers)

        integrals_R = []
        for i in boundary_conditions:
            if 'Robin' in boundary_conditions[i]:
                r, s = boundary_conditions[i]['Robin']
                integrals_R.append(r*(self.u-s)*self.v*ds(i))

        self.constraints += sum(integrals_R)

        # Number conservation constraints

        N0 = self.L_scaled*self.c_scaled  # total amount of species in cell
        for k in range(self.M):
            self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
                'Ion species {:02d} number conservation constraint'.format(k),
                N0[k], lwidth=self.label_width))
            self.applyNumberConservationConstraint(k, self.c_scaled[k])

    def discretize(self):
        """Builds function space, call again after introducing constraints"""
        # FEniCS interface
        self.mesh = fn.IntervalMesh(self.N, self.x0_scaled, self.x1_scaled)

        # http://www.femtable.org/
        # Argyris*                          ARG
        # Arnold-Winther*                   AW
        # Brezzi-Douglas-Fortin-Marini*     BDFM
        # Brezzi-Douglas-Marini             BDM
        # Bubble                            B
        # Crouzeix-Raviart                  CR
        # Discontinuous Lagrange            DG
        # Discontinuous Raviart-Thomas      DRT
        # Hermite*                          HER
        # Lagrange                          CG
        # Mardal-Tai-Winther*               MTW
        # Morley*                           MOR
        # Nedelec 1st kind H(curl)          N1curl
        # Nedelec 2nd kind H(curl)          N2curl
        # Quadrature                        Q
        # Raviart-Thomas                    RT
        # Real                              R

        # construct test and trial function space from elements
        # spanned by Lagrange polynomials for the pyhsical variables of
        # potential and concentration and global elements with a single degree
        # of freedom ('Real') for constraints.
        # For an example of this approach, refer to
        #     https://fenicsproject.org/docs/dolfin/latest/python/demos/neumann-poisson/demo_neumann-poisson.py.html
        # For another example on how to construct and split function spaces
        # for solving coupled equations, refer to
        #     https://fenicsproject.org/docs/dolfin/latest/python/demos/mixed-poisson/demo_mixed-poisson.py.html

        P = fn.FiniteElement('Lagrange', fn.interval, 3)
        R = fn.FiniteElement('Real', fn.interval, 0)
        elements = [P]*(1+self.M) + [R]*self.K

        H = fn.MixedElement(elements)
        self.W = fn.FunctionSpace(self.mesh, H)

        # solution functions
        self.w = fn.Function(self.W)

        # set initial values if available
        P = fn.FunctionSpace(self.mesh, 'P', 1)
        dof2vtx = fn.vertex_to_dof_map(P)
        if self.ui0 is not None:
            x = np.linspace(self.x0_scaled, self.x1_scaled, self.ui0.shape[0])
            ui0 = scipy.interpolate.interp1d(x, self.ui0)
            # use linear interpolation on mesh
            self.u0_func = fn.Function(P)
            self.u0_func.vector()[:] = ui0(self.X)[dof2vtx]
            fn.assign(self.w.sub(0),
                      fn.interpolate(self.u0_func, self.W.sub(0).collapse()))

        if self.ni0 is not None:
            x = np.linspace(self.x0_scaled, self.x1_scaled, self.ni0.shape[1])
            ni0 = scipy.interpolate.interp1d(x, self.ni0)
            self.p0_func = [fn.Function(P)]*self.ni0.shape[0]
            for k in range(self.ni0.shape[0]):
                self.p0_func[k].vector()[:] = ni0(self.X)[k, :][dof2vtx]
                fn.assign(self.w.sub(1+k),
                          fn.interpolate(self.p0_func[k], self.W.sub(k+1).collapse()))

        # u represents voltage , p concentrations
        uplam = fn.split(self.w)
        self.u, self.p, self.lam = (
            uplam[0], [*uplam[1:(self.M+1)]], [*uplam[(self.M+1):]])

        # v, q and mu represent respective test functions
        vqmu = fn.TestFunctions(self.W)
        self.v, self.q, self.mu = (
            vqmu[0], [*vqmu[1:(self.M+1)]], [*vqmu[(self.M+1):]])

    def __init__(self, *args, **kwargs):
        self.init(*args, **kwargs)

        # TODO: don't hard code btcol
        self.bctol = 1e-14  # tolerance for identifying domain boundaries
        self.K = 0  # number of Lagrange multipliers (constraints)
        self.constraints = 0  # holds constraint kernels
        self.discretize()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""
Applies steric correction to coordiante sample.

Copyright 2020 IMTEK Simulation
University of Freiburg

Authors:

    Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>

Examples:

    Apply one steric radius for all species:

        stericify.py --verbose --radii 2.0 -- input.lammps output.lammps

    Apply per-species steric radii (for a system of supposedly two species):

        stericify.py --verbose --radii 2.0 5.0 -- input.lammps output.lammps

    Overrides other parameters otherwise infered from input file:

        stericify.py --verbose --box 48 48 196 --names Na Cl --charges 1 -1 \\
            --radii 2.0 5.0 -- input.lammps output.lammps
"""
import logging
import os
import sys
import time

import ase
import ase.io
import numpy as np

try:
    import json
    import urllib.parse
except ImportError:
    pass

from matscipy.electrochemistry.steric_correction import apply_steric_correction
from matscipy.electrochemistry.steric_correction import scipy_distance_based_closest_pair


def main():
    """Applies steric correction to coordiante sample. Assures a certain
    minimum pairwiese distance between points in sample and
    between points and box boundary.

    ATTENTION: LAMMPS data file export (atom style 'full') requires ase>3.20.0
    """
    logger = logging.getLogger()

    import argparse

    # in order to have both:
    # * preformatted help text and ...
    # * automatic display of defaults
    class ArgumentDefaultsAndRawDescriptionHelpFormatter(
            argparse.ArgumentDefaultsHelpFormatter,
            argparse.RawDescriptionHelpFormatter):
        pass

    class StoreAsNumpyArray(argparse._StoreAction):
        def __call__(self, parser, namespace, values, option_string=None):
            values = np.array(values, ndmin=1)
            return super().__call__(
                parser, namespace, values, option_string)

    class StoreAsDict(argparse._StoreAction):
        def __call__(self, parser, namespace, value, option_string=None):
            if 'json' not in sys.modules or 'urllib' not in sys.modules:
                raise ModuleNotFoundError(
                    "Modules 'json' and 'urllib' required for parsing dicts.")
            try:
                parsed_value = json.loads(urllib.parse.unquote(value))
            except json.decoder.JSONDecodeError as exc:
                logger.error("Failed parsing '{}'".format(value))
                raise exc

            return super().__call__(
                parser, namespace, parsed_value, option_string)

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)

    parser.add_argument('infile', metavar='IN', nargs='?',
                        help='.xyz or .lammps (LAMMPS data) format input file')
    parser.add_argument('outfile', metavar='OUT', nargs='?',
                        help='.xyz or .lammps (LAMMPS data) format output file')

    parser.add_argument('--radii', '-r', default=[2.0], type=float, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('R'), required=False, dest="radii",
                        help=('Steric radii, either one for all or '
                              'species-wise. Same units as distances in input.'))

    parser.add_argument('--box', '-b', default=None, nargs=3,
                        action=StoreAsNumpyArray,
                        metavar=('X', 'Y', 'Z'), required=False, type=float,
                        dest="box", help=('Bounding box, overrides cell from'
                                          'input. Same units as distances in input.'))

    parser.add_argument('--names', default=None, type=str, nargs='+',
                        metavar=('NAME'), required=False, dest="names",
                        help='Atom names, overrides names from input')

    parser.add_argument('--charges', default=None, type=float, nargs='+',
                        action=StoreAsNumpyArray,
                        metavar=('Z'), required=False, dest="charges",
                        help='Atom charges, overrides charges from input')

    parser.add_argument('--method', type=str,
                        metavar=('METHOD'), required=False,
                        dest="method",
                        default='L-BFGS-B',
                        help='Scipy minimizer')

    parser.add_argument('--options', type=str,
                        action=StoreAsDict,
                        metavar=('JSON DICT'), required=False,
                        dest="options",
                        default={
                            'gtol':    1.e-12,
                            'ftol':    1.e-12,
                            'maxiter': 100,
                            'disp':    False,
                        },
                        help=(
                            'Convergence options for scipy minimier.'
                            ' Pass as JSON-formatted key:value dict. See'
                            ' https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html'
                            ' for minimizer-specific options.'))

    parser.add_argument('--debug', default=False, required=False,
                        action='store_true', dest="debug", help='debug flag')
    parser.add_argument('--verbose', default=False, required=False,
                        action='store_true', dest="verbose",
                        help='verbose flag')
    parser.add_argument('--log', required=False, nargs='?', dest="log",
                        default=None, const='c2d.log', metavar='LOG',
                        help=(
                            'Write log file c2d.log, optionally specify log'
                            ' file name'))

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
        # This supports bash autocompletion.
        # To enable this, 'pip install argcomplete',
        # then activate global completion.
    except ImportError:
        pass

    args = parser.parse_args()

    if args.debug:
        loglevel = logging.DEBUG
    elif args.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    # logformat  = ''.join(("%(asctime)s",
    #  "[ %(filename)s:%(lineno)s - %(funcName)s() ]: %(message)s"))
    logformat = "[ %(filename)s:%(lineno)s - %(funcName)s() ]: %(message)s"

    logging.basicConfig(level=loglevel,
                        format=logformat)

    # explicitly modify the root logger (necessary?)
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # remove all handlers
    for h in logger.handlers:
        logger.removeHandler(h)

    # create and append custom handles
    ch = logging.StreamHandler()
    formatter = logging.Formatter(logformat)
    ch.setFormatter(formatter)
    ch.setLevel(loglevel)
    logger.addHandler(ch)

    if args.log:
        fh = logging.FileHandler(args.log)
        fh.setFormatter(formatter)
        fh.setLevel(loglevel)
        logger.addHandler(fh)

    logger.info('This is `{}` : `{}`.'.format(__file__, __name__))

    logger.debug('Args: {}'.format(args))

    # input validation
    if not args.infile:
        infile = sys.stdin
        infile_format = 'xyz'
    else:
        infile = args.infile
        _, infile_format = os.path.splitext(infile)

    logger.info('Input format {} from {}.'.format(infile_format, infile))

    if infile_format == '.lammps':
        system = ase.io.read(
            infile, format='lammps-data', units="real", style='full')
    else:  # elif outfile_format == '.xyz'
        system = ase.io.read(infile, format='extxyz')

    n = len(system)  # total number of particles
    logger.info('Read "{}" system within bounding box'.format(system.symbols))
    for l in system.cell:
        logger.info('    [ {:> 8.2e},{:> 8.2e},{:> 8.2e} ]'.format(*l))

    species_atomic_numbers = np.unique(system.get_atomic_numbers())
    species_symbols = [
        ase.data.chemical_symbols[i] for i in species_atomic_numbers]

    n_species = len(species_atomic_numbers)
    logger.info('    containing {:d} particles of {:d} species'.format(
        n, n_species))
    if not isinstance(args.radii, np.ndarray):
        args.radii = np.array(args.radii, ndmin=1)

    r = np.zeros(n)
    if len(args.radii) == 1:
        logger.info('Applying steric radius r = {:.2e} to all species.'.format(
            args.radii[0]))
        r[:] = args.radii[0]

    elif len(args.radii) == n_species:
        for i, (a, s) in enumerate(zip(species_atomic_numbers, species_symbols)):
            logger.info(
                'Applying steric radius r = {:.2e} for species {:s}.'.format(
                    args.radii[i], s))
            r[system.get_atomic_numbers() == a] = args.radii[i]
    else:
        raise ValueError(
            """Steric radii must either be one value for all species or one value
            per species, i.e. {:d} values in your case.""".format(n_species))

    if args.box is not None:
        if not isinstance(args.box, np.ndarray):
            args.box = np.array(args.box, ndmin=1)

        logger.info('Box specified on command line')
        logger.info('   [ {:> 8.2e},{:> 8.2e},{:> 8.2e} ]'.format(*args.box))
        logger.info('overides box from input data.')
        system.set_cell(args.box)

    if args.charges is not None:
        logger.info('Charges specified on command line reassign input charges to')
        new_charges = np.ndarray(n, dtype=int)

        for a, s, c in zip(species_atomic_numbers, species_symbols, args.charges):
            logger.info(
                '    {:s} -> {}'.format(s, c))
            new_charges[system.get_atomic_numbers() == a] = c
        system.set_initial_charges(new_charges)

    if args.names is not None:
        new_species_symbols = args.names
        new_species_atomic_numbers = [
            ase.data.atomic_numbers[name] for name in new_species_symbols]
        logger.info('Species specified on command line reassign input species to')
        new_atomic_numbers = np.ndarray(n, dtype=int)
        for aold, anew, sold, snew in zip(species_atomic_numbers,
                                          new_species_atomic_numbers,
                                          species_symbols, new_species_symbols):
            logger.info(
                '    {:s} -> {:s}'.format(sold, snew))

            new_atomic_numbers[system.get_atomic_numbers() == aold] = anew
        system.set_atomic_numbers(new_atomic_numbers)
        species_atomic_numbers = new_species_atomic_numbers
        species_symbols = new_species_symbols

        specorder = args.names  # assure type ordering as specified on cmdline
    else:
        specorder = None  # keep ordering as is

    # prepare for minimization
    x0 = system.get_positions()

    # only works for orthogonal box
    box3 = np.array(system.get_cell_lengths_and_angles())[0:3]
    box6 = np.array([[0., 0., 0], box3])  # needs lower corner

    # n = x0.shape[0], set above
    # dim = x0.shape[1]

    # benchmakr methods
    mindsq, (p1, p2) = scipy_distance_based_closest_pair(x0)
    pmin = np.min(x0, axis=0)
    pmax = np.max(x0, axis=0)
    mind = np.sqrt(mindsq)
    logger.info("Minimum pair-wise distance in initial sample: {}".format(mind))
    logger.info("First sample point in pair:    ({:8.4e},{:8.4e},{:8.4e})".format(*p1))
    logger.info("Second sample point in pair    ({:8.4e},{:8.4e},{:8.4e})".format(*p2))
    logger.info("Box lower boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box6[0]))
    logger.info("Minimum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmin))
    logger.info("Maximum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmax))
    logger.info("Box upper boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box6[1]))

    t0 = time.perf_counter()
    x1, res = apply_steric_correction(x0, box=box6, r=r,
                                      method=args.method, options=args.options)
    # use default method and options

    t1 = time.perf_counter()
    dt = t1 - t0
    logger.info("{} s runtime".format(dt))

    mindsq, (p1, p2) = scipy_distance_based_closest_pair(x1)
    mind = np.sqrt(mindsq)
    pmin = np.min(x1, axis=0)
    pmax = np.max(x1, axis=0)

    logger.info("Finished with status = {}, success = {}, #it = {}".format(
        res.status, res.success, res.nit))
    logger.info("    message = '{}'".format(res.message))
    logger.info("Minimum pair-wise distance in final configuration: {:8.4e}".format(mind))
    logger.info("First sample point in pair:    ({:8.4e},{:8.4e},{:8.4e})".format(*p1))
    logger.info("Second sample point in pair    ({:8.4e},{:8.4e},{:8.4e})".format(*p2))
    logger.info("Box lower boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box6[0]))
    logger.info("Minimum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmin))
    logger.info("Maximum coordinates in sample: ({:8.4e},{:8.4e},{:8.4e})".format(*pmax))
    logger.info("Box upper boundary:            ({:8.4e},{:8.4e},{:8.4e})".format(*box6[1]))

    diff = x1 - x0
    n_diff = np.count_nonzero(diff)
    diffnorm = np.linalg.norm(diff)
    logger.info(
      '{:d} coords. differ numerically in final and initial config.'.format(
        n_diff
      ))
    logger.info('Norm of difference between final and initial config')
    logger.info('    || x1 - x0 || = {:.4e}'.format(diffnorm))

    system.set_positions(x1)

    if not args.outfile:
        outfile = sys.stdout
        outfile_format = '.xyz'
    else:
        outfile = args.outfile
        _, outfile_format = os.path.splitext(outfile)

    logger.info('Output format {} to {}.'.format(outfile_format, outfile))

    if outfile_format == '.lammps':
        ase.io.write(
            outfile, system,
            format='lammps-data', units="real", atom_style='full',
            specorder=specorder)
    else:  # elif outfile_format == '.xyz'
        ase.io.write(outfile, system, format='extxyz')

    logger.info('Done.')


if __name__ == '__main__':
    # Execute everything else
    main()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#                  Adrien Gola, Karlsruhe Institute of Technology
#                  Wolfram Nöhring, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""Read and write tabulated EAM potentials"""

from collections import namedtuple

import numpy as np
try:
    from scipy import interpolate
except:
    print('Warning: No scipy')
    interpolate = False

import os

from ase.units import Hartree, Bohr

###


"""Conversion factor from Hartree to Electronvolt. \
Use 1 Hartree = 27.2 eV to be consistent with Lammps EAM."""
_hartree_in_electronvolt = 27.2

"""Conversion factor from Bohr radii to Angstrom. \
Use 1 Bohr radius = 0.529 A to be consistent with Lammps EAM."""
_bohr_in_angstrom = 0.529

"""Conversion factor for charge function of EAM potentials \
of kind 'eam' (DYNAMO funcfl format)."""
_hartree_bohr_in_electronvolt_angstrom = _hartree_in_electronvolt * _bohr_in_angstrom

# Todo: replace by data class (requires Python > 3.7)
class EAMParameters(
    namedtuple(
        "EAMParameters",
        "symbols atomic_numbers "
        "atomic_masses lattice_constants crystal_structures "
        "number_of_density_grid_points "
        "number_of_distance_grid_points "
        "density_grid_spacing distance_grid_spacing "
        "cutoff",
    )
):
    """Embedded Atom Method potential parameters

    :param array_like symbols: Symbols of the elements coverered by
        this potential (only for eam/alloy and 
        eam/fs, EMPTY for eam
    :param array_like atomic_numbers: Atomic numbers of the elements 
        covered by this potential
    :param array_like atomic_masses: Atomic masses of the elements 
        covered by this potential
    :param array_like lattice_constants: Lattice constant of a pure crystal 
        with crystal structure as specified in crystal_structures
    :param array_like crystal_structures: Crystal structure of the pure metal.
    :param int number_of_density_grid_points: Number of grid points 
        of the embedding energy functional
    :param int number_of_distance_grid_points: Number of grid points of 
        the electron density function and the pair potential
    :param float density_grid_spacing: Grid spacing in electron density space
    :param float distance_grid_spacing: Grid spacing in pair distance space
    :param float cutoff: Cutoff distance of the potential
    """

    __slots__ = ()

###

def _strip_comments_from_line(string, marker="#"):
    """Strip comments from lines but retain newlines

    Parameters
    ----------
    string : str
        string which may contain comments
    marker : str
        marker which indicates the start of a comment

    Returns
    -------
    stripped_string : str
        string without commments; if the string terminated
        with a newline, then the newline is retained
    """
    start = string.find(marker)
    if start != -1:
        stripped_string = string[:start]
        if string.endswith("\n"):
            stripped_string += "\n"
    else:
        stripped_string = string
    return stripped_string


def read_eam(eam_file, kind="eam/alloy"):
    """Read a tabulated EAM potential
    
    There are differnt flavors of EAM, with different storage
    formats. This function supports a subset of the formats supported
    by Lammps (http://lammps.sandia.gov/doc/pair_eam.html),
    * eam (DYNAMO funcfl format)
    * eam/alloy (DYNAMO setfl format)
    * eam/fs (DYNAMO setfl format)
    
    Parameters
    ----------
    eam_file : string
        eam alloy file name 
    kind : {'eam', 'eam/alloy', 'eam/fs'}
        kind of EAM file to read

    Returns
    -------
    source : string
        Source informations or comment line for the file header
    parameters : EAMParameters
        EAM potential parameters
    F : array_like
        contain the tabulated values of the embedded functions
        shape = (nb elements, nb of data points)
    f : array_like
        contain the tabulated values of the density functions
        shape = (nb elements, nb of data points)
    rep : array_like
        contain the tabulated values of pair potential
        shape = (nb elements,nb elements, nb of data points)
    """
    supported_kinds = ["eam", "eam/alloy", "eam/fs"]
    if kind not in supported_kinds: 
        raise ValueError(f"EAM kind {kind} not supported")
    with open(eam_file, 'r') as file:
        eam = file.readlines()

    if kind == "eam":
        with open(eam_file, 'r') as file:
            # ignore comment characters on first line but strip them from subsequent lines
            lines = [file.readline()]
            lines.extend(_strip_comments_from_line(line) for line in file.readlines())
        # reading first comment line as source for eam potential data
        source = lines[0].strip()

        words = lines[1].strip().split()
        if len(words) != 4:
            raise ValueError(
                "expected four values on second line of EAM setfl file: "
                "atomic number, mass, lattice constant, lattice"
            )
        # Turn atomic numbers and masses, lattice parameter, and crystal
        # structures into arrays to be consistent with the other EAM styles
        atomic_numbers = np.array((int(words[0]), ), dtype=int)
        atomic_masses = np.array((float(words[1]), ), dtype=float)
        lattice_parameters = np.array((float(words[2]),), dtype=float)
        crystal_structures = np.empty(1).astype(np.str)
        crystal_structures[0] = words[3] 

        words = lines[2].strip().split()
        if len(words) != 5:
            raise ValueError(
                "expected five values on third line of EAM setfl file: "
                "Nrho, drho, Nr, dr, cutoff"
            )
        Nrho = int(words[0])       # Nrho (number of values for the embedding function F(rho))
        drho = float(words[1])     # spacing in density space
        Nr = int(words[2])         # Nr (number of values for the effective charge function Z(r) and density function rho(r))
        dr = float(words[3])       # spacing in distance space
        cutoff = float(words[4])
        parameters = EAMParameters(
            np.zeros(1), atomic_numbers, atomic_masses, lattice_parameters, 
            crystal_structures, Nrho, Nr, drho, dr, cutoff
        )

        # Strip empty lines 
        remaining_lines = [line for line in lines[3:] if len(line.strip()) > 0]
        remaining_words = []
        for line in remaining_lines:
            words = line.split()
            remaining_words.extend(words)
        expected_length = Nrho + 2 * Nr
        true_length = len(remaining_words)
        if true_length != expected_length:
            raise ValueError(f"expected {expected_length} tabulated values, but there are {true_length}")
        data = np.array(remaining_words, dtype=float)
        F = data[0:Nrho]
        # 'eam' (DYNAMO funcfl) tables contain the charge function :math:`Z`,
        # and not the pair potential :math:`\phi`. :math:`Z` needs to be 
        # converted into :math:`\phi` first, which involves unit conversion.
        # To be consistent with the other eam styles (and avoid complications
        # later), we convert into :math:`r*\phi`, where :math:`r` is the pair distance, i.e.
        # r = np.arange(0, rep.size) * dr
        charge = data[Nrho:Nrho+Nr]
        rep = charge**2
        rep *= _hartree_bohr_in_electronvolt_angstrom 
        f = data[Nrho+Nr:2*Nr+Nrho]
        # Reshape in order to be consistent with other EAM styles
        return source, parameters, F.reshape(1, Nrho), f.reshape(1, Nr), rep.reshape(1, 1, Nr)

    if kind in ["eam/alloy", "eam/fs"]:
        """eam/alloy and eam/fs have almost the same structure, except for the electron density section"""
        with open(eam_file, 'r') as file:
            # ignore comment characters on first line but strip them from subsequent lines
            lines = [file.readline() for _ in range(3)]
            lines.extend(_strip_comments_from_line(line) for line in file.readlines())
        # reading 3 first comment lines as source for eam potential data
        source = "".join(line.strip() for line in lines[:3])

        words = lines[3].strip().split()
        alleged_num_elements = int(words[0])
        elements = words[1:]
        true_num_elements = len(elements)
        if alleged_num_elements != true_num_elements:
            raise ValueError(
                f"Header claims there are tables for {alleged_num_elements} elements, "
                f"but actual element list has {true_num_elements} elements: {' '.join(elements)}"
            )

        words = lines[4].strip().split()
        Nrho = int(words[0])     # Nrho (number of values for the embedding function F(rho))
        drho = float(words[1])   # spacing in density space
        Nr = int(words[2])       # Nr (number of values for the effective charge function Z(r) and density function rho(r))
        dr = float(words[3])     # spacing in distance space
        cutoff = float(words[4])

        # Strip empty lines and check that the table contains the expected number of values
        remaining_lines = [line for line in lines[5:] if len(line.strip()) > 0]
        remaining_words = []
        for line in remaining_lines:
            words = line.split()
            remaining_words.extend(words)
        if kind == "eam/fs":
            expected_num_density_functions_per_element = true_num_elements
        else:
            expected_num_density_functions_per_element = 1
        expected_num_words_per_element = (
            4 + 
            Nrho + 
            expected_num_density_functions_per_element * Nr 
        )
        expected_num_pair_functions = np.sum(np.arange(1, true_num_elements+1)).astype(int)
        expected_length = true_num_elements * expected_num_words_per_element + expected_num_pair_functions * Nr
        true_length = len(remaining_words)
        if true_length != expected_length:
            raise ValueError(f"expected {expected_length} tabulated values, but there are {true_length}")

        atomic_numbers = np.zeros(true_num_elements, dtype=int)
        atomic_masses = np.zeros(true_num_elements)
        lattice_parameters = np.zeros(true_num_elements)
        crystal_structures = np.empty(true_num_elements).astype(np.str) # fixme: be careful with string length
        F = np.zeros((true_num_elements, Nrho))
        for i in range(true_num_elements):
            offset = i * expected_num_words_per_element
            atomic_numbers[i] = int(remaining_words[offset])
            atomic_masses[i] = float(remaining_words[offset+1])
            lattice_parameters[i] = float(remaining_words[offset+2])
            crystal_structures[i] = remaining_words[offset+3]
            F[i, :] = np.array(remaining_words[offset+4:offset+4+Nrho], dtype=float)

        # Read data for individual elemements
        if kind == "eam/alloy":
            f = np.zeros((true_num_elements, Nr))
            for i in range(true_num_elements):
                offset = i * expected_num_words_per_element + 4 + Nrho
                f[i, :] = np.array(remaining_words[offset:offset+Nr], dtype=float)
        if kind == "eam/fs":
            f = np.zeros((true_num_elements, true_num_elements, Nr))
            for i in range(true_num_elements):
                offset = i * expected_num_words_per_element + 4 + Nrho
                for j in range(true_num_elements):
                    f[i, j, :] = np.array(remaining_words[offset+j*Nr:offset+(j+1)*Nr], dtype=float)

        # Read pair data
        rep = np.zeros((true_num_elements, true_num_elements, Nr))
        rows, cols = np.tril_indices(true_num_elements)
        for pair_number, (i, j) in enumerate(zip(rows, cols)):
            offset = true_num_elements * expected_num_words_per_element + pair_number * Nr
            rep[i, j, :] = np.array(remaining_words[offset:offset+Nr], dtype=float)
            rep[j, i, :] = rep[i, j, :]

        parameters = EAMParameters(
            elements, atomic_numbers, atomic_masses, 
            lattice_parameters, crystal_structures, 
            Nrho, Nr, drho, dr, cutoff
        )
        return source, parameters, F, f, rep
            

def mix_eam(files,kind,method,f=[],rep_ab=[],alphas=[],betas=[]):
    """
    mix eam alloy files data set and compute the interspecies pair potential part using the 
    mean geometric value from each pure species 
    
    Parameters
    ----------
    files : array of strings
            Contain all the files to merge and mix
    kind : string
            kinf of eam. Supported eam/alloy, eam/fs
    method : string, {geometric, arithmetic, weighted, fitted}
        Method used to mix the pair interaction terms. The geometric,
        arithmetic, and weighted arithmetic average are available. The weighted
        arithmetic method is using the electron density function values of atom
        :code:`a` and :code:`b` to ponderate the pair potential between species
        :code:`a` and :math:`b`, :code:`rep_ab = 0.5(fb/fa * rep_a + fa/fb *
        rep_b)`, see [1]. The fitted method is to be used if :code:`rep_ab`
        has been previously fitted and is parse as :math:`rep_ab` karg.
    f : np.array 
        fitted density term (for FS eam style)
    rep_ab : np.array 
        fitted rep_ab term
    alphas : array
        fitted alpha values for the fine tuned mixing. 
        :code:`rep_ab = alpha_a*rep_a+alpha_b*rep_b`
    betas : array 
        fitted values for the fine tuned mixing. 
        :code:`f_ab = beta_00*rep_a+beta_01*rep_b`
        :code:`f_ba = beta_10*rep_a+beta_11*rep_b`

    Returns
    -------
    sources : string
        Source informations or comment line for the file header
    parameters_mix: EAMParameters
        EAM potential parameters
    F_ : array_like
        contain the tabulated values of the embedded functions
        shape = (nb elements, nb elements, nb of data points)
    f_ : array_like
        contain the tabulated values of the density functions
        shape = (nb elements, nb elements, nb of data points)
    rep_ : array_like
        contain the tabulated values of pair potential
        shape = (nb elements, nb elements, nb of data points)

    References
    ----------

    1. X. W. Zhou, R. A. Johnson, and H. N. G. Wadley, Phys. Rev. B, 69, 144113 (2004)
    """

    nb_at = 0
    # Counting elements and repartition and select smallest tabulated set Nrho*drho // Nr*dr
    Nrho,drho,Nr,dr,cutoff = np.empty((len(files))),np.empty((len(files))),np.empty((len(files))),np.empty((len(files))),np.empty((len(files)))
    sources = ""
    if kind == "eam/alloy":
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            sources+= source
            source += " "
            nb_at+=len(parameters[0])
            Nrho[i] = parameters[5]
            drho[i] = parameters[7]
            cutoff[i] = parameters[9]
            Nr[i] = parameters[6]
            dr[i] = parameters[8]
        # --- #
        max_cutoff = cutoff.argmax()
        max_prod = (Nrho*drho).argmax()
        max_prod_r = (Nr*dr).argmax()
        atomic_numbers,atomic_masses,lattice_parameters,crystal_structures,elements = np.empty(0),np.empty(0),np.empty(0),np.empty(0).astype(np.str),np.empty(0).astype(np.str)
        Nr_ = Nr[max_prod_r]
        dr_ = ((Nr*dr).max())/Nr_
        Nrho_ = Nrho[max_prod]
        drho_ = ((Nrho*drho).max())/Nrho_
        
        if Nr_ > 2000:
          Nr_ = 2000   # reduce
          dr_ = ((Nr*dr).max())/Nr_
        if Nrho_ > 2000:
          Nrho_ = 2000 # reduce
          drho_ = ((Nrho*drho).max())/Nrho_
        F_,f_,rep_ = np.empty((nb_at,Nrho_)),np.empty((nb_at,Nr_)),np.empty((nb_at,nb_at,Nr_))
        at = 0
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            elements = np.append(elements,parameters[0])
            atomic_numbers = np.append(atomic_numbers,parameters[1])
            atomic_masses = np.append(atomic_masses,parameters[2])
            lattice_parameters = np.append(lattice_parameters,parameters[3])
            crystal_structures = np.append(crystal_structures,parameters[4])
            for j in range(len(parameters[0])):
                F_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nrho[i]*drho[i],Nrho[i]),F[j,:])(np.linspace(0,Nrho_*drho_,Nrho_))
                f_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),f[j,:])(np.linspace(0,Nr_*dr_,Nr_))
                rep_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),rep[j,j,:])(np.linspace(0,Nr_*dr_,Nr_))
                at+=1
        # mixing repulsive part
        old_err_state = np.seterr(divide='raise')
        ignored_states = np.seterr(**old_err_state)
        for i in range(nb_at):
            for j in range(nb_at):
                if j < i :
                    if method == "geometric":
                        rep_[i,j,:] = (rep_[i,i,:]*rep_[j,j,:])**0.5
                    if method == "arithmetic":
                        if alphas:
                            rep_[i,j,:] = alphas[i]*rep_[i,i,:]+alphas[j]*rep_[j,j,:]
                        else:
                            rep_[i,j,:] = 0.5*(rep_[i,i,:]+rep_[j,j,:])
                    if method == "weighted":
                        rep_[i,j,:] = 0.5*(np.divide(f_[j,:],f_[i,:])*rep_[i,i,:]+np.divide(f_[i,:],f_[j,:])*rep_[j,j,:])
                    if method == "fitted":
                        rep_ab[np.isnan(rep_ab)] = 0
                        rep_ab[np.isinf(rep_ab)] = 0
                        rep_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    rep_[i,j,:][np.isnan(rep_[i,j,:])] = 0
                    rep_[i,j,:][np.isinf(rep_[i,j,:])] = 0
    elif kind == "eam/fs":
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            sources+= source
            source += " "
            nb_at+=len(parameters[0])
            Nrho[i] = parameters[5]
            drho[i] = parameters[7]
            cutoff[i] = parameters[9]
            Nr[i] = parameters[6]
            dr[i] = parameters[8]
        # --- #
        max_cutoff = cutoff.argmax()
        max_prod = (Nrho*drho).argmax()
        max_prod_r = (Nr*dr).argmax()
        atomic_numbers,atomic_masses,lattice_parameters,crystal_structures,elements = np.empty(0),np.empty(0),np.empty(0),np.empty(0).astype(np.str),np.empty(0).astype(np.str)
        Nr_ = Nr[max_prod_r]
        dr_ = ((Nr*dr).max())/Nr_
        Nrho_ = Nrho[max_prod]
        drho_ = ((Nrho*drho).max())/Nrho_
        
        if Nr_ > 2000:
          Nr_ = 2000   # reduce
          dr_ = ((Nr*dr).max())/Nr_
        if Nrho_ > 2000:
          Nrho_ = 2000 # reduce
          drho_ = ((Nrho*drho).max())/Nrho_
        F_,f_,rep_ = np.empty((nb_at,Nrho_)),np.empty((nb_at,nb_at,Nr_)),np.empty((nb_at,nb_at,Nr_))
        at = 0
        for i,f_eam in enumerate(files):
            source,parameters, F,f,rep = read_eam(f_eam,kind="eam/alloy")
            elements = np.append(elements,parameters[0])
            atomic_numbers = np.append(atomic_numbers,parameters[1])
            atomic_masses = np.append(atomic_masses,parameters[2])
            lattice_parameters = np.append(lattice_parameters,parameters[3])
            crystal_structures = np.append(crystal_structures,parameters[4])
            for j in range(len(parameters[0])):
                F_[at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nrho[i]*drho[i],Nrho[i]),F[j,:])(np.linspace(0,Nrho_*drho_,Nrho_))
                f_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),f[j,:])(np.linspace(0,Nr_*dr_,Nr_))
                rep_[at,at,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,Nr[i]*dr[i],Nr[i]),rep[j,j,:])(np.linspace(0,Nr_*dr_,Nr_))
                at+=1
        # mixing density part
        old_err_state = np.seterr(divide='raise')
        ignored_states = np.seterr(**old_err_state)
        for i in range(nb_at):
            for j in range(nb_at):
                if i!=j:
                    if method == "geometric":
                        f_[i,j,:] = (f_[i,i,:]*f_[j,j,:])**0.5
                    if method == "arithmetic":
                        if betas.any():
                            f_[i,j,:] = betas[i,i]*f_[i,i,:]+betas[i,j]*f_[j,j,:]
                        else:
                            f_[i,j,:] = 0.5*(f_[i,i,:]+f_[j,j,:])
                    if method == "fitted":
                        f_ab[np.isnan(f_ab)] = 0
                        f_ab[np.isinf(f_ab)] = 0
                        f_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    f_[i,j,:][np.isnan(f_[i,j,:])] = 0
                    f_[i,j,:][np.isinf(f_[i,j,:])] = 0
        # mixing repulsive part
        for i in range(nb_at):
            for j in range(nb_at):
                if j < i :
                    if method == "geometric":
                        rep_[i,j,:] = (rep_[i,i,:]*rep_[j,j,:])**0.5
                    if method == "arithmetic":
                        if alphas:
                            rep_[i,j,:] = alphas[i]*rep_[i,i,:]+alphas[j]*rep_[j,j,:]
                        else:
                            rep_[i,j,:] = 0.5*(rep_[i,i,:]+rep_[j,j,:])
                    if method == "fitted":
                        rep_ab[np.isnan(rep_ab)] = 0
                        rep_ab[np.isinf(rep_ab)] = 0
                        rep_[i,j,:] = interpolate.InterpolatedUnivariateSpline(np.linspace(0,max(Nr*dr),rep_ab.shape[0]),rep_ab)(np.linspace(0,Nr_*dr_,Nr_))
                    rep_[i,j,:][np.isnan(rep_[i,j,:])] = 0
                    rep_[i,j,:][np.isinf(rep_[i,j,:])] = 0
    else:
        raise ValueError(f"EAM kind {kind} is not supported")
                
    parameters_mix = EAMParameters(elements, atomic_numbers, atomic_masses,lattice_parameters,crystal_structures, Nrho_,Nr_, drho_, dr_, cutoff[max_cutoff])
    return sources, parameters_mix, F_, f_, rep_

      
def write_eam(source, parameters, F, f, rep, out_file, kind="eam"):
    """Write an eam lammps format file 

    There are differnt flavors of EAM, with different storage
    formats. This function supports a subset of the formats supported
    by Lammps (http://lammps.sandia.gov/doc/pair_eam.html),
    * eam (DYNAMO funcfl format)
    * eam/alloy (DYNAMO setfl format)
    * eam/fs (DYNAMO setfl format)
    
    Parameters
    ----------
    source : string
          Source information or comment line for the file header
    parameters_mix: EAMParameters
        EAM potential parameters
    F : array_like
        contain the tabulated values of the embedded functions
        shape = (nb of data points)
    f : array_like
        contain the tabulated values of the density functions
        shape = (nb of data points)
    rep : array_like
        contain the tabulated values of pair potential
        shape = (nb of data points)
    out_file : string
              output file name for the eam alloy potential file
    kind : {'eam', 'eam/alloy', 'eam/fs'}
        kind of EAM file to read

    Returns
    -------
    None
    """
  
    elements, atomic_numbers, atomic_masses, lattice_parameters, crystal_structures = parameters[0:5]
    Nrho, Nr, drho, dr, cutoff = parameters[5:10]
    
    if kind == "eam":
        # parameters unpacked
        # FIXME: atomic numbers etc are now arrays, and not scalars
        crystal_structures_str = ' '.join(s for s in crystal_structures)
        atline = f"{int(atomic_numbers)} {float(atomic_masses)}  {float(lattice_parameters)} {crystal_structures_str}"
        parameterline = f'{int(Nrho)}\t{float(drho):.16e}\t{int(Nr)}\t{float(dr):.16e}\t{float(cutoff):.10e}'
        potheader = f"# EAM potential from : # {source} \n {atline} \n {parameterline}"
        # --- Writing new EAM alloy pot file --- #
        # write header and file parameters
        potfile = open(out_file,'wb')
        # write F and pair charge tables
        Nr = parameters.number_of_distance_grid_points
        Nrho = parameters.number_of_density_grid_points
        np.savetxt(potfile, F.reshape(Nrho), fmt='%.16e', header=potheader, comments='')
        # 'eam' (DYNAMO funcfl) tables contain the charge function :math:`Z`,
        # and not the pair potential :math:`\phi`. The array :code:`rep` 
        # stores :math:`r*\phi`, where :math:`r` is the pair distance, 
        # and we can convert using the relation :math:`r\phi=z*z`, where
        # additional unit conversion to units of sqrt(Hartree*Bohr-radii)
        # is required.
        charge = rep / _hartree_bohr_in_electronvolt_angstrom 
        charge = np.sqrt(charge)
        np.savetxt(potfile, charge.reshape(Nr), fmt='%.16e')
        # write electron density tables
        np.savetxt(potfile, f.reshape(Nr), fmt='%.16e')
        potfile.close()  
    elif kind == "eam/alloy":
        num_elements = len(elements)
        # parameters unpacked
        potheader = f"# Mixed EAM alloy potential from :\n# {source} \n# \n"
        # --- Writing new EAM alloy pot file --- #
        potfile = open(out_file,'wb')
        # write header and file parameters
        np.savetxt(
            potfile, elements, fmt="%s", newline=' ', 
            header=potheader+str(num_elements), 
            footer=f'\n{Nrho}\t{drho:e}\t{Nr}\t{dr:e}\t{cutoff:e}\n', 
            comments=''
        )
        # write F and f tables
        for i in range(num_elements):
            np.savetxt(
                potfile, np.append(F[i,:], f[i,:]), fmt="%.16e", 
                header=f'{atomic_numbers[i]:d}\t{atomic_masses[i]}\t{lattice_parameters[i]}\t{crystal_structures[i]}',
                comments=''
            )
        # write pair interactions tables
        [[np.savetxt(potfile,rep[i,j,:],fmt="%.16e") for j in range(rep.shape[0]) if j <= i] for i in range(rep.shape[0])]
        potfile.close() 
    elif kind == "eam/fs":
        num_elements = len(elements)
        # parameters unpacked
        potheader = f"# Mixed EAM fs potential from :\n# {source} \n# \n"
        # --- Writing new EAM alloy pot file --- #
        potfile = open(out_file,'wb')
        # write header and file parameters
        np.savetxt(
            potfile, elements, fmt="%s", newline=' ', 
            header=potheader+str(num_elements), 
            footer=f'\n{Nrho}\t{drho:e}\t{Nr}\t{dr:e}\t{cutoff:e}\n', 
            comments=''
        )
        # write F and f tables
        for i in range(num_elements):
            np.savetxt(
                potfile, np.append(F[i,:], f[i,:,:].flatten()), fmt="%.16e", 
                header=f'{atomic_numbers[i]:d}\t{atomic_masses[i]}\t{lattice_parameters[i]}\t{crystal_structures[i]}',
                comments=''
            )
        # write pair interactions tables
        [[np.savetxt(potfile, rep[i,j,:], fmt="%.16e") for j in range(rep.shape[0]) if j <= i] for i in range(rep.shape[0])]
        potfile.close() 
    else:
        raise ValueError(f"EAM kind {kind} is not supported")
